/*
 * Hibernate, Relational Persistence for Idiomatic Java
 *
 * JBoss, Home of Professional Open Source
 * Copyright 2015 Red Hat Inc. and/or its affiliates and other contributors
 * as indicated by the @authors tag. All rights reserved.
 * See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 *
 * This copyrighted material is made available to anyone wishing to use,
 * modify, copy, or redistribute it subject to the terms and conditions
 * of the GNU Lesser General Public License, v. 2.1.
 * This program is distributed in the hope that it will be useful, but WITHOUT A
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License,
 * v.2.1 along with this distribution; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301, USA.
 */
package org.hibernate.search.indexes.serialization.avro.impl;

import org.apache.avro.Protocol;
import org.hibernate.search.indexes.serialization.avro.logging.impl.Log;
import org.hibernate.search.util.logging.impl.LoggerFactory;

/**
 * Keeps a reference to each known Avro Protocol revision we can support.
 * Only a single Major version is supported, but for minor revisions we
 * should always be able to deserialize an older encoding scheme.
 * For minor revisions beyond the last known version a best effort is
 * applied and a warning is logged.
 * Each Protocol is only created at first use, so to not waste parsing
 * time nor memory for unused revisions.
 *
 * @author Sanne Grinovero
 * @since 5.0
 */
final class KnownProtocols {

	/**
	 * Latest protocol version is 1.1
	 */
	static final int MAJOR_VERSION = 1;
	static final int LATEST_MINOR_VERSION = 1;

	private static final Log log = LoggerFactory.make( Log.class );
	private volatile Protocol v1_0 = null;
	private volatile Protocol v1_1 = null;
	private volatile boolean warned = false;

	Protocol getProtocol(int majorVersion, int minorVersion) {
		if ( MAJOR_VERSION != majorVersion ) {
			throw log.incompatibleProtocolVersion(
					majorVersion,
					minorVersion,
					MAJOR_VERSION,
					LATEST_MINOR_VERSION
			);
		}
		if ( minorVersion == 1 ) {
			return getV1_1();
		}
		else if ( minorVersion == 0 ) {
			return getV1_0();
		}
		else {
			if ( ! warned ) {
				warned = true;
				log.unexpectedMinorProtocolVersion( majorVersion, minorVersion, LATEST_MINOR_VERSION );
			}
			return getV1_1();
		}
	}

	Protocol getLatestProtocol() {
		return getProtocol( MAJOR_VERSION, LATEST_MINOR_VERSION );
	}

	private Protocol getV1_0() {
		Protocol p = v1_0;
		if ( p == null ) {
			synchronized (this) {
				p = v1_0;
				if ( p != null ) {
					return p;
				}
				p = new ProtocolBuilderV1_0().build();
				v1_0 = p;
			}
		}
		return p;
	}

	private Protocol getV1_1() {
		Protocol p = v1_1;
		if ( p == null ) {
			synchronized (this) {
				p = v1_1;
				if ( p != null ) {
					return p;
				}
				p = new ProtocolBuilderV1_1().build();
				v1_1 = p;
			}
		}
		return p;
	}

}
